/*
 * LinkedList header
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1 $  $Date: 2003/01/11 12:40:09 $
 * ---------------------------------------------------------------------------
 */
#ifndef _XL_LINKED_LIST
# define _XL_LINKED_LIST

#include <phlib/xl_config.h>

/* Make sure NULL is defined */
#ifndef NULL
#  define NULL 0
#endif

struct _xl_ll_node;
typedef struct _xl_ll_node Node;

struct _xl_ll_node {
  Node *next;
  Node *prev;
  int  nodesize;
  uchar_t *ptr;
};


struct _xl_ll_list {
  Node *head;      /* Pointer to the head of the list  */
  Node *last;      /* Pointer to LAST node in list     */
  Node *current;   /* The last node accessed           */
  int   datasize;  /* Size of user data stored in list */
  int   nodesize;  /* size of whole node inc user data */
  int   count;     /* Number of nodes in list          */
};

typedef struct _xl_ll_list LIST;

typedef signed int ListPos;
#define ListStart (-1)
#define ListEnd   (-2)

#define xl_List_insert_node(a,b,c) xl_List_insert_a_node((a),(b),(c),1)
#define xl_List_insert_space(a,b)  xl_List_insert_a_node((a),(b),0,0)

/* Prototypes */
RetValue xl_List_init        ( LIST **listref, int user_data_size );
RetValue xl_List_end         ( LIST **listref );
RetValue xl_List_transfer_node( LIST *listref, LIST *src, ListPos pos );
RetValue xl_List_insert_a_node( LIST *listref, ListPos pos, 
				uchar_t *userdata, int set );
RetValue xl_List_insert_nodeS( LIST *listref, ListPos pos, 
			       uchar_t *userdata, int uds );
RetValue xl_List_delete_node ( LIST *listref );
RetValue xl_List_make_first  ( LIST *listref );
RetValue xl_List_get_node    ( LIST *listref, ListPos pos, uchar_t **userdata );
RetValue xl_List_get_first   ( LIST *listref, uchar_t **userdata );
RetValue xl_List_get_last    ( LIST *listref, uchar_t **userdata );
RetValue xl_List_get_next    ( LIST *listref, uchar_t **userdata );
RetValue xl_List_get_prev    ( LIST *listref, uchar_t **userdata );
int      xl_List_set_datasize( LIST *listref, int user_data_size );
int      xl_List_get_size    ( LIST *listref );
int      xl_List_get_nodesize( LIST *listref );
char *   xl_List_get_current_data( LIST *listref );

#endif

/* ---------------------------------------------------------------------- */
/* end */
