/*
 * Code to handle files
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:16:00 $
 * ---------------------------------------------------------------------- 
 */

#include <phlib/xl_config.h>
#include <sys/stat.h>
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#include <fcntl.h>

#include <errno.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>

/* ----------------------------------------------------------------------- */

RetValue
xl_flock( int fdesc )
{
  RetValue ret = Success;
  struct flock lock_info; 
  
  /* Lock file */
  lock_info.l_type   = F_WRLCK;
  lock_info.l_whence = 0;
  lock_info.l_start  = 0;
  lock_info.l_len    = 0;
  
  if( fcntl(fdesc, F_SETLK, &lock_info) == MINUSONE ) {
    /* failed to lock */
    ret = Failure;
  }
  
  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
xl_funlock(int fdesc)
{
  RetValue ret = Success;
  struct flock lock_info; 
  
  /* Unlock file */
  lock_info.l_type = F_UNLCK;
  lock_info.l_whence = 0;
  lock_info.l_start = 0;
  lock_info.l_len = 0;
  
  fcntl(fdesc, F_SETLK, &lock_info);
  
  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
xl_wait_on_lock( int fd, int secs )
{
  RetValue ret = Success;
  struct flock lock_info; 
  int i = 0;

  lock_info.l_type = F_WRLCK;
  lock_info.l_whence = 0;
  lock_info.l_start = 0;
  lock_info.l_len = 0;
  
  fcntl( fd, F_GETLK, &lock_info );

  while( lock_info.l_type == F_WRLCK && i++ < secs ) sleep(1);
  if( i >= secs ) ret = Failure;

  return ret;
}


/* ----------------------------------------------------------------------- */
/* end */
