/*
 * Really Scary Constant Memory Buffer Linked List Implementation.
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.1.1.1 $  $Date: 2003/01/10 23:16:02 $
 * ---------------------------------------------------------------------------
 * 
 * No pointers are used and so this linked list may be written to
 * disk, shared by multiple processes in SHM or whatever....
 * - Disadvantages is that it requires a fixed memory block to work in and
 *    cannot grow outside this data size....
 * - Advantages are that it's quite fast for this reason (not as fast as a
 *    pooled xl_linklst setup though!)
 *
 * No animals were used in the testing of this code :o?
 * ---------------------------------------------------------------------------
 */
#include <stdlib.h>
#include <string.h>
#include <phlib/xl_types.h>
#include <phlib/xl_memlst.h>

#define NODE_PTR(a,b) (SNode*)((char*)(a)+(a)->o_qroot+((b)*(a)->blocksize))
#define WORD_SIZE 4

static unsigned int val_2_bit[] = {
  0,
  0,
  1, 0,
  2, 0,0,0,
  3, 0,0,0,0,0,0,0,
  4, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  5, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  6, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  7
};

/* Private prototypes */
static RetValue shmp_create_node ( int *node,    SHMLIST *listref );
static void     shmp_release_node( int oldblock, SHMLIST *listref );

/* ---------------------------------------------------------------------- */
/*
 * Initialise a linked list.
 * You pass a pointer to a great big data area and it'll use what it needs...
 * This is a bit nasty of course, so an additional function is provided
 *   which returns the amount of space that _would_ be required, so 
 *   provision for it can be made in advance.
 */
RetValue
shm_List_init( SHMLIST *list, int user_data_size, int maxnodes  )
{
  int idxsize = WORD_SIZE * 8;
  int blksize = WORD_SIZE;

  list->mapsize = ((maxnodes / idxsize) + 1) * blksize; /* Mapsize in bytes */

  list->o_qmap  = sizeof( SHMLIST );
  list->o_qroot = sizeof( SHMLIST ) + list->mapsize;

  list->maxnodes   = maxnodes;
  list->bo_head    = -1;
  list->bo_last    = -1;
  list->bo_current = -1;
  list->nodesize   = user_data_size + sizeof( SNode );
  list->datasize   = user_data_size;  /* Set to max - overridden in insert */
  list->count      = 0;

  /* Block is rounded to complete word chunks */
  list->blocksize  = list->nodesize + 
    ( WORD_SIZE - list->nodesize % WORD_SIZE ) % WORD_SIZE;

  /* Setup block map (set all bits to 1 meaning all free) */
  memset( (char*)list + list->o_qmap, 0xff, list->mapsize );

  return Success;
}


/* ---------------------------------------------------------------------- */

int
shm_List_determine_memory_req( int userdata, int maxnodes )
{
  int a;
  /* calculate extra bytes required to round to the next whole wordsize */
  a = ( WORD_SIZE - ( userdata + sizeof(SNode) ) % WORD_SIZE ) % WORD_SIZE;

  /*
    printf( 
    "User Data: %d\n"
    "  Rounded: %d\n", userdata, (WORD_SIZE-(userdata + sizeof(SNode))%WORD_SIZE) % WORD_SIZE );

    printf( "Single block size: %ld\n", userdata +  a );
  */

  return
    sizeof( SHMLIST ) +                            /* Control Data size */
    (userdata + sizeof(SNode) + a) * maxnodes +     /* Space for nodes   */
    ((maxnodes / (WORD_SIZE*8)) + 1) * WORD_SIZE;  /* Space for NodeMap */
}

/* ---------------------------------------------------------------------- */
/*
 * Free up a linked list
 */ 
RetValue
shm_List_end( SHMLIST **listref )
{
  /* TODO: ditch list */
  return Success;
}

/* ---------------------------------------------------------------------- */
/*
 * THIS IS CALLED VIA MACROS - shm_List_insert_node and ...insert_space
 *
 * This function inserts nodes into the linked list.
 * It will insert at the Start, End or somewhere in the middle.....
 * If the insert position indicator is greater then the number of nodes
 * in the list then a Failure return will occur.
 * It only transfers the number of bytes defined as the blocksize when the
 * list was created!!!!
 */
RetValue
shm_List_insert_a_node( SHMLIST *listref, ListPos pos, 
			uchar *userdata, int set, int len )
{
  int  new, keep;
  SNode *nptr = NULL;
  SNode *kptr = NULL;
  RetValue ret = Failure;
  ListPos  mpos = pos;
  int i;

  ASSERT( listref );

  if( len > listref->datasize ) return Failure;
  if( listref->count >= listref->maxnodes ) return Failure;

  /* Fixup incase list is empty or we're on the last node */
  if( mpos == listref->count ) mpos = ListEnd;
  if( listref->count == 0 )    mpos = 0;

  switch( mpos ) {
  case 0:
  case ListStart: 
    if( shmp_create_node( &new, listref ) == Success ) {
      keep = listref->bo_head;
      listref->bo_head = new;

      if( keep != -1 ) (NODE_PTR( listref, keep ))->Bprev = new;
      nptr = NODE_PTR( listref, new );

      nptr->Bnext = keep;
      nptr->Bprev = -1;
      mpos        = 0;
      ret = Success;
    }
    break;

  case ListEnd:
    if( shmp_create_node( &new, listref ) == Success ) {

      nptr = NODE_PTR( listref, new );

      nptr->Bprev = listref->bo_last;
      nptr->Bnext = -1;
      (NODE_PTR( listref, nptr->Bprev ))->Bnext = new;
      
      listref->bo_last = new; /* We know we're the new end tag.... */
      ret = Success;
    }
    break;

  default:
    if( mpos <= listref->count && mpos > 0 ) {
      if( shmp_create_node( &new, listref ) == Success ) {

	keep = listref->bo_head; /* Point at first node */
	for( i = 0; i < mpos; i++ ) /* Find require node */
	  keep = (NODE_PTR( listref, keep ))->Bnext;

	nptr = NODE_PTR( listref, new );
	kptr = NODE_PTR( listref, keep );

	nptr->Bprev       = kptr->Bprev;

	(NODE_PTR( listref, kptr->Bprev ))->Bnext = new;

	nptr->Bnext      = keep;
	kptr->Bprev      = new;

	ret = Success;
      }
    }
    break;
  }

  if( ret == Success ) {
    if( mpos == listref->count ) listref->bo_last = new; /* If new end tag */
    listref->count++;
    listref->bo_current = new;          /* Record last node accessed */
    nptr->nodesize = listref->nodesize; /* Record size of node       */
    nptr->o_data   = sizeof( SNode );

    if(set == 1) {
      memcpy( (char*)nptr + nptr->o_data, userdata, len );
      nptr->datalen = len;
    } else nptr->datalen = 0;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

char *
shm_List_get_current_data( SHMLIST *listref )
{
  SNode *nptr = NODE_PTR( listref, listref->bo_current );

  return (char*)nptr + nptr->o_data;
}

/* ---------------------------------------------------------------------- */

int
shm_List_get_size( SHMLIST *listref )
{
  ASSERT( listref );
  return listref->count;
}

/* ---------------------------------------------------------------------- */

int
shm_List_get_nodesize( SHMLIST *listref )
{
  ASSERT( listref );

  if( listref->bo_current == -1 ) return (-1);

  return (NODE_PTR( listref, listref->bo_current )) -> nodesize;
}

/* ---------------------------------------------------------------------- */

RetValue
shm_List_get_node( SHMLIST *listref, ListPos pos, uchar **userdata )
{
  RetValue ret = Failure;
  int node;
  int i;

  ASSERT( listref );

  if( pos < listref->count ) {
    node = listref->bo_head; /* Point at first node */
    for( i = 0; i < pos; i++ ) 
      node = (NODE_PTR( listref, node ))->Bnext; /* Find req. */

    listref->bo_current = node;     /* Last node accessed */
    *userdata = (uchar*)(NODE_PTR( listref, node )) + 
      (NODE_PTR( listref, node ))->o_data;
    ret = Success;
  }
  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * This deletes the current list node.
 * If the node is in the middle of the list (ie not first or last) then
 * the new 'current' node becomes the one on the left.
 * However, if we are deleteing the first node, then the new first node
 * becomes the node to the right and also the current node!
 */
RetValue
shm_List_delete_node( SHMLIST *listref )
{
  RetValue ret = Failure;
  int right, kill, left;

  ASSERT( listref );
  kill = listref->bo_current;

  if( kill != -1 ) {
    right = (NODE_PTR( listref, kill ))->Bnext;
    left  = (NODE_PTR( listref, kill ))->Bprev;

    shmp_release_node( kill, listref );
    
    /* If there is no right node then we're deleting the last node */
    if( right == -1 ) {
      listref->bo_last = left; /* New last node */
    } else {
      (NODE_PTR(listref, right ))->Bprev = left;
    }
    
    if( left != -1 ) {
      (NODE_PTR( listref, left ))->Bnext = right;
      listref->bo_current = left;  /* The left one is the last accessed.. */
    } else {
      /* Hmm, there is no left node! */
      listref->bo_head    = right;
      listref->bo_current = right; /* Of course.. this may be NULL!       */
    }

    listref->count--;
      
#if DEBUG
    if( listref->count < 0 ) {
      printf( "Error!  Can't have a negative list item count!\n" );
      exit( 1 ) ;
    }
#endif
    
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
shm_List_get_first( SHMLIST *listref, uchar **userdata )
{
  RetValue ret = Failure;
  int node;

  ASSERT( listref );
  node = listref->bo_head;          /* Point to first cell */
  if( node != -1 ) {
    listref->bo_current = node;     /* Last node accessed */
    *userdata = (uchar*)(NODE_PTR( listref, node )) + 
      (NODE_PTR( listref, node ))->o_data;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
shm_List_get_last( SHMLIST *listref, uchar **userdata )
{
  RetValue ret = Failure;
  int node;

  ASSERT( listref );
  node = listref->bo_last;          /* Point to last cell */
  if( node != -1 ) {
    listref->bo_current = node;     /* Last node accessed */
    *userdata = (uchar*)(NODE_PTR( listref, node )) + 
      (NODE_PTR( listref, node ))->o_data;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
shm_List_get_next( SHMLIST *listref, uchar **userdata )
{
  RetValue ret = Failure;
  int node;

  ASSERT( listref );
  node = listref->bo_current;       /* Point to last node accessed */

  if( node != -1 && (NODE_PTR( listref, node ))->Bnext != -1 ) {

    node = (NODE_PTR( listref, node ))->Bnext;/* Step forward to next node   */
    listref->bo_current = node;               /* Remember last node accessed */

    *userdata = (uchar*)(NODE_PTR( listref, node )) + 
      (NODE_PTR( listref, node ))->o_data;
    ret = Success;
  }

  return ret;
}

/* ---------------------------------------------------------------------- */

RetValue
shm_List_get_prev( SHMLIST *listref, uchar **userdata )
{
  RetValue ret = Failure;
  int node;

  ASSERT( listref );
  node = listref->bo_current;       /* Point to last node accessed */

  if( node != -1 && (NODE_PTR( listref, node ))->Bprev != -1 ) {

    node = (NODE_PTR( listref, node))->Bprev; /* Step back to last node     */
    listref->bo_current = node;              /* Remember last node accessed */

    *userdata = (uchar*)(NODE_PTR( listref, node )) + 
      (NODE_PTR( listref, node ))->o_data;
    ret = Success;
  }

  return ret;
}
/* ---------------------------------------------------------------------- */
/*
 * This PRIVATE function locates a free 'block' in our storage area
 * and marks it as being used.  The block number is returned.
 */
static RetValue
shmp_create_node( int *newblock, SHMLIST *listref )
{
  RetValue ret = Failure;
  unsigned int *map;
  unsigned int i, b, w, t;

  map = (unsigned int*)( (char*)listref + listref->o_qmap );
  i = listref->mapsize * 8;
  b = 0;
  while( b < i ) {
    w = *map;
    t = w & -w;
    if( t ) {
      /* Clear bit, indicating that it's now being used */
      *map &= ~t;
  
      /* work out the bit number (block number) and return it */
      if     ( t > 0x00ffffff ) b += val_2_bit[ t >> 24 ] + 24;
      else if( t > 0x0000ffff ) b += val_2_bit[ t >> 16 ] + 16;
      else if( t > 0x000000ff ) b += val_2_bit[ t >> 8  ] + 8;
      else b += val_2_bit[ t ];
      ret = Success;
      break;
    }
    b += 32;
    map++;
  }

#if DEBUG
  printf( "NODE: %d\n", ret == Success ? b:-1 );
#endif

  *newblock = ret == Success ? b:-1;
  return ret;
}

/* ---------------------------------------------------------------------- */
/*
 * Set a bit in the block map indicating that it is free
 */
static void
shmp_release_node( int b, SHMLIST *listref )
{
  ((int*)( (char*)listref + listref->o_qmap ))[ b/32 ] |= 1 << (b%32);
}

/* ---------------------------------------------------------------------- */

#ifdef DEBUG

static RetValue
debug_map( int *firstfree, SHMLIST *listref )
{
  RetValue ret = Failure;
  unsigned int *map;
  unsigned int i, b, w, t;
  map = (unsigned int*)( (char*)listref + listref->o_qmap );
  i = listref->mapsize * 8;

  b = 0;
  while( b < i ) {
    w = *map;
    t = w & -w;
    if( t ) {
      /* work out the bit number (block number) and return it */
      if     ( t > 0x00ffffff ) b += val_2_bit[ t >> 24 ] + 24;
      else if( t > 0x0000ffff ) b += val_2_bit[ t >> 16 ] + 16;
      else if( t > 0x000000ff ) b += val_2_bit[ t >> 8  ] + 8;
      else b += val_2_bit[ t ];
      ret = Success;
      break;
    }
    b += 32;
    map++;
  }
  
  *firstfree = ret == Success ? b:-1;
  return ret;
}


#define DATASIZE 13

void
testnodes()
{


  int size, i;
  SHMLIST *list;
  int items = 400;
  int free;
  uchar buffer[ DATASIZE + 1 ];

  size = shm_List_determine_memory_req( DATASIZE, items );

  printf( "Size required: %d\n", size );

  list = (SHMLIST*)malloc( size );

  /* memory, datasize, number of blocks */
  shm_List_init( list, DATASIZE, items  );

  for( i = 0; i < 50; i++ ) {
    debug_map( &free, list );
    printf( "First Freeblock: %d\n", free );
    shm_List_insert_node(list, ListEnd, buffer );
    
    if( i > 25 ) {
      printf( "   delete node %d\n", i - i/2 );
      shm_List_get_node( list, i - i/2, (uchar**)&buffer );
      shm_List_delete_node( list );
      printf( "\n" );
    }
  }
}

#endif

/* ---------------------------------------------------------------------- */
/* end */
